<?php

/**
 * Provides helper functions related to DB tables, created by custom modules.
 */
class CustomSchemaHelper {

  /**
   * Contains db tables provided by custom modules.
   */
  private array $schema = [];

  /**
   * Class constructor.
   */
  public function __construct() {
    $this->collectSchema();
  }

  /**
   * Returns custom tables.
   *
   * @return array
   *   The list of DB table names provided by custom modules.
   */
  public function getTables(): array {
    return $this->schema;
  }

  /**
   * Collects data related to DB tables provided by custom modules.
   *
   * The code is inspired by drupal_get_complete_schema().
   */
  private function collectSchema(): void {
    // Load the .install files to get hook_schema.
    if (function_exists('module_load_all_includes')) {
      module_load_all_includes('install');
    }

    // Invoke hook_schema for all modules.
    foreach (module_implements('schema') as $moduleName) {
      if (!$this->isModuleCustom($moduleName)) {
        continue;
      }

      $moduleSchema = (array) module_invoke($moduleName, 'schema');
      $this->schema = array_merge($this->schema, array_keys($moduleSchema));
    }
  }

  /**
   * Checks if module is a custom module.
   *
   * @param string $moduleName
   *   The module name.
   *
   * @return bool
   *   TRUE if module is located in a folder like .../modules/custom,
   *   or FALSE otherwise.
   */
  private function isModuleCustom(string $moduleName): bool {
    $modulePath = drupal_get_path('module', $moduleName);
    return strpos($modulePath, 'modules/custom') > 0
      || strpos($modulePath, 'modules/feature') > 0
      || strpos($modulePath, 'modules/third') > 0;
  }

}
