<?php

/**
 * Generates a config for DB Anonymizer.
 */
class DbaConfigGenerator {

  /**
   * Custom schema helper.
   *
   * @var \CustomSchemaHelper
   */
  private $customSchemaHelper;

  /**
   * Contains generated site config.
   */
  private array $config;

  /**
   * The prefix of drupal tables in DB.
   */
  private string $tablePrefix = '';

  /**
   * Percentage of data to dump for entity types.
   */
  private array $dataPercentage = [
    'default' => 0.01,
    'entity_info' => -1,
    'field_collection_item' => 0,
    'file' => 0,
    'taxonomy_term' => 1,
    'taxonomy_vocabulary' => -1,
    'wysiwyg_profile' => -1,
  ];

  /**
   * Class constructor.
   */
  public function __construct() {
    $this->customSchemaHelper = new CustomSchemaHelper();
  }

  /**
   * Generates site config for DB Anonymizer.
   *
   * @return array
   *   Site config (to be exported as yaml).
   */
  public function generate(): array {
    $this->generateConfigSkeleton();
    $this->generateConfigTables();
    $this->generateConfigEntities();

    return $this->config;
  }

  /**
   * Generates site config skeleton.
   */
  private function generateConfigSkeleton(): void {
    global $databases;

    $this->tablePrefix = $databases['default']['default']['prefix'];
    $this->config = [
      'name' => variable_get('site_name'),
      'driver' => $databases['default']['default']['driver'],
      'database' => [],
    ];

    if ($customTables = $this->customSchemaHelper->getTables()) {
      $this->config['database']['__READ_AND_REMOVE_ME__'] = [
        'description' => 'There are some tables provided by custom modules have been found. Please check these tables now as you may want to add them as relations to some entities. Once you are done remove this section from the config please.',
        'custom_tables' => $customTables,
      ];
    }
  }

  /**
   * Generates config for some tables export.
   */
  private function generateConfigTables(): void {
    $this->config['database']['tables'][$this->table('batch')] = 0;
    $this->config['database']['tables'][$this->table('batch_job')] = 0;
    $this->config['database']['tables'][$this->table('queues')] = 0;
    $this->config['database']['tables'][$this->table('sessions')] = 0;
    $this->config['database']['tables'][$this->table('watchdog')] = 0;
    $this->config['database']['tables']['cache_'] = [
      'table_regex' => '/^' . $this->table('cache') . '.*/',
      'get' => 0,
    ];
    $this->config['database']['tables']['field_deleted_data_'] = [
      'table_regex' => '/^' . $this->table('field_deleted_data') . '.*/',
      'get' => 0,
    ];
    $this->config['database']['tables']['field_deleted_revision_'] = [
      'table_regex' => '/^' . $this->table('field_deleted_revision') . '.*/',
      'get' => 0,
    ];
    $this->config['database']['tables']['field_revision_'] = [
      'table_regex' => '/^' . $this->table('field_revision') . '.*/',
      'get' => 0,
    ];
    $this->config['database']['tables']['search_'] = [
      'table_regex' => '/^' . $this->table('search_') . '.*/',
      'get' => 0,
    ];
  }

  /**
   * Generates config for drupal entities.
   */
  private function generateConfigEntities(): void {
    $entityTypes = entity_get_info();
    ksort($entityTypes);

    // Go through all entity types.
    foreach ($entityTypes as $entityTypeId => $entityType) {
      // Call custom entity type anonymization config generator.
      $method = 'generateAnonymizationConfig' . str_replace('_', '', ucwords($entityTypeId, '_'));
      if (method_exists($this, $method)) {
        $this->{$method}($entityTypes, $entityTypeId);
      }

      // Call custom entity type config generator (if exists).
      $method = 'generateConfigEntity' . str_replace('_', '', ucwords($entityTypeId, '_'));
      if (method_exists($this, $method)) {
        $this->{$method}($entityTypes, $entityTypeId);
      }
      elseif ($this->shouldExportEntity($entityTypeId)) {
        $this->generateConfigEntity($entityTypes, $entityTypeId);
      }
    }
  }

  /**
   * Generates config for Taxonomy Term entity type.
   *
   * @param array $entityTypes
   *   Entity types data.
   * @param string $entityTypeId
   *   Current entity type id.
   */
  private function generateConfigEntityTaxonomyTerm(array $entityTypes, string $entityTypeId): void {
    $entityType = $entityTypes[$entityTypeId];

    // Get taxonomy vocabularies in [machine_name => vid] format.
    $vocabularies = taxonomy_vocabulary_load_multiple(FALSE);
    $vocabularies = array_combine(
      array_column($vocabularies, 'machine_name'),
      array_column($vocabularies, 'vid')
    );

    foreach ($entityType['bundles'] ?? [] as $bundleId => $bundle) {
      $entityConfig = [
        'table' => $this->table($entityType['base table']),
        'get' => $this->getPercentage($entityTypeId),
        'where' => [
          'vid' => $vocabularies[$bundleId],
        ],
        'fields' => [
          'id' => $entityType['entity keys']['id'],
        ],
      ];

      $this->appendEntityFieldsConfig($entityTypes, $entityTypeId, $bundleId, $entityConfig);
      $this->appendEntityExtraRelationsConfig($entityTypes, $entityTypeId, $bundleId, $entityConfig);
      $this->config['database']['entities'][$entityTypeId . '__' . $vocabularies[$bundleId]] = $entityConfig;
    }
  }

  /**
   * Generates config for File entity type.
   *
   * @param array $entityTypes
   *   Entity types data.
   * @param string $entityTypeId
   *   Current entity type id.
   */
  private function generateConfigEntityFile(array $entityTypes, string $entityTypeId): void {
    $entityType = $entityTypes[$entityTypeId];

    $entityConfig = [
      'get' => $this->getPercentage($entityTypeId),
      'fields' => [
        'id' => $entityType['entity keys']['id'],
      ],
    ];
    $entityConfig['relations'][$this->table('file_usage')] = [
      'where' => [
        $entityType['entity keys']['id'] => '%' . $entityType['entity keys']['id'],
      ],
    ];

    $this->config['database']['entities'][$this->table($entityType['base table'])] = $entityConfig;
  }

  /**
   * Generates config for User entity type.
   *
   * @param array $entityTypes
   *   Entity types data.
   * @param string $entityTypeId
   *   Current entity type id.
   */
  private function generateConfigEntityUser(array $entityTypes, string $entityTypeId): void {
    $entityType = $entityTypes[$entityTypeId];

    // Generate config to always have users with uid=0 and uid=1.
    $entityConfig = $this->generateConfigEntity($entityTypes, $entityTypeId, [
      'get' => 1,
      'where' => [$entityType['entity keys']['id'] => [1, '<=']],
      'where__no_bundle' => TRUE,
      'do_not_add' => TRUE,
    ]);
    $this->config['database']['entities'][$entityTypeId . '__drupal'] = $entityConfig;

    // Also append config to get normal percentage of random users.
    $entityConfig['get'] = $this->getPercentage($entityTypeId);
    $entityConfig['where'] = [$entityType['entity keys']['id'] => [1, '>']];
    $this->config['database']['entities'][$entityTypeId] = $entityConfig;
  }

  /**
   * This is generic method to generate config for some entity type.
   *
   * @param array $entityTypes
   *   Entity types data.
   * @param string $entityTypeId
   *   Current entity type id.
   * @param array $params
   *   Parameters for entity config.
   *
   * @return array
   *   Generated config for the last entity bundle.
   */
  private function generateConfigEntity(array $entityTypes, string $entityTypeId, array $params = []): array {
    $entityType = $entityTypes[$entityTypeId];

    foreach ($entityType['bundles'] ?? [] as $bundleId => $bundle) {
      $entityConfig = [
        'table' => $this->table($entityType['base table']),
        'get' => $params['get'] ?? $this->getPercentage($entityTypeId),
      ];

      if (!empty($params['where'])) {
        $entityConfig['where'] = $params['where'];
      }
      if (empty($params['where__no_bundle'])) {
        $entityConfig['where'][$entityType['entity keys']['bundle']] = $bundleId;
      }

      $entityConfig['fields']['id'] = $entityType['entity keys']['id'];

      $this->appendEntityFieldsConfig($entityTypes, $entityTypeId, $bundleId, $entityConfig);
      $this->appendEntityExtraRelationsConfig($entityTypes, $entityTypeId, $bundleId, $entityConfig);

      if (empty($params['do_not_add'])) {
        $this->config['database']['entities'][$entityTypeId . '__' . $bundleId] = $entityConfig;
      }
    }

    return $entityConfig ?? [];
  }

  /**
   * Adds field config as relations to entity config.
   *
   * @param array $entityTypes
   *   Entity types data.
   * @param string $entityTypeId
   *   Current entity type id.
   * @param string $entityBundleId
   *   Current entity bundle id.
   * @param array $entityConfig
   *   Current entity type/bundle config. Field config is added to this config.
   */
  private function appendEntityFieldsConfig(array $entityTypes, string $entityTypeId, string $entityBundleId, array &$entityConfig): void {
    $fields = field_info_instances($entityTypeId, $entityBundleId);
    ksort($fields);

    // Go through all entity fields.
    foreach ($fields as $fieldName => $data) {
      $field = field_info_field($fieldName);
      $storage = $field['storage']['details']['sql'];
      $tableWithoutPrefix = array_keys($storage['FIELD_LOAD_CURRENT'])[0];
      $table = $this->table($tableWithoutPrefix);

      $entityConfig['relations'][$table] = [
        'where' => [
          'entity_type' => $entityTypeId,
          'bundle' => $entityBundleId,
          'entity_id' => '%' . $entityTypes[$entityTypeId]['entity keys']['id'],
        ],
      ];

      // If field contains just 1 value then add limit to boost
      // the search query.
      if (($field['cardinality'] ?? 0) == 1) {
        $entityConfig['relations'][$table]['limit'] = 1;
      }

      // Check if a field is a reference to another entity.
      $referencedEntityTypeId = strpos($field['type'], '_reference')
        ? str_replace('_reference', '', $field['type'])
        : '';

      if (!$referencedEntityTypeId) {
        if ($field['type'] === 'field_collection') {
          $referencedEntityTypeId = 'field_collection_item';
        }
        elseif ($field['type'] === 'file') {
          $referencedEntityTypeId = 'file';
        }
      }

      // If a field is an entity reference then we need to add a relation
      // to dump related entity as well.
      if ($referencedEntityTypeId && $this->shouldExportEntity($referencedEntityTypeId)) {
        $foreignIdFieldName = !empty($field['foreign keys'])
          ? array_keys($field['foreign keys'])[0]
          : '';
        $entityConfig['relations'][$table]['relations'][$fieldName] = [
          'table' => $this->table($entityTypes[$referencedEntityTypeId]['base table']),
          'is_entity' => 1,
          'values' => [
            'id' => '%' . $storage['FIELD_LOAD_CURRENT'][$tableWithoutPrefix][array_keys($storage['FIELD_LOAD_CURRENT'][$tableWithoutPrefix])[0]],
            'type' => $referencedEntityTypeId === 'file'
              ? $field['foreign keys'][$foreignIdFieldName]['table']
              : $referencedEntityTypeId,
          ],
          'fields' => [
            'id' => $entityTypes[$referencedEntityTypeId]['entity keys']['id'],
            'bundle' => $referencedEntityTypeId === 'taxonomy_term' ? 'vid' : $entityTypes[$referencedEntityTypeId]['entity keys']['bundle'],
          ],
        ];
      }
    }
  }

  /**
   * Adds extra relations config to entity config.
   *
   * @param array $entityTypes
   *   Entity types data.
   * @param string $entityTypeId
   *   Current entity type id.
   * @param string $entityBundleId
   *   Current entity bundle id.
   * @param array $entityConfig
   *   Current entity type/bundle config. Relations are added to this config.
   */
  private function appendEntityExtraRelationsConfig(array $entityTypes, string $entityTypeId, string $entityBundleId, array &$entityConfig): void {
    $relations = $this->getEntityExtraRelations($entityTypes, $entityTypeId);
    foreach ($relations as $table => $config) {
      $entityConfig['relations'][$this->table($table)] = $config;
    }
  }

  /**
   * Returns extra relations for the entity type.
   *
   * Extra relations are tables related to the entity type, besides the
   * field tables.
   *
   * @param array $entityTypes
   *   Entity types data.
   * @param string $entityTypeId
   *   Current entity type id.
   *
   * @return array
   *   Config for extra relations that can be appended to entity type config.
   */
  private function getEntityExtraRelations(array $entityTypes, string $entityTypeId): array {
    $relations = [];

    switch ($entityTypeId) {
      case 'node':
        $relations['node_revision'] = [
          'where' => [
            $entityTypes[$entityTypeId]['entity keys']['id'] => '%' . $entityTypes[$entityTypeId]['entity keys']['id'],
            $entityTypes[$entityTypeId]['entity keys']['revision'] => '%' . $entityTypes[$entityTypeId]['entity keys']['revision'],
          ],
        ];
        $relations['node_access'] = [
          'where' => [
            $entityTypes[$entityTypeId]['entity keys']['id'] => '%' . $entityTypes[$entityTypeId]['entity keys']['id'],
          ],
        ];
        $relations['taxonomy_index'] = [
          'where' => [
            $entityTypes[$entityTypeId]['entity keys']['id'] => '%' . $entityTypes[$entityTypeId]['entity keys']['id'],
          ],
        ];
        break;

      case 'taxonomy_term':
        $relations['taxonomy_term_hierarchy'] = [
          'where' => [
            $entityTypes[$entityTypeId]['entity keys']['id'] => '%' . $entityTypes[$entityTypeId]['entity keys']['id'],
          ],
        ];
        break;

      case 'user':
        $relations['users_roles'] = [
          'where' => [
            $entityTypes[$entityTypeId]['entity keys']['id'] => '%' . $entityTypes[$entityTypeId]['entity keys']['id'],
          ],
        ];
        break;
    }

    return $relations;
  }

  /**
   * Generates config for File entity table anonymization.
   *
   * @param array $entityTypes
   *   Entity types data.
   * @param string $entityTypeId
   *   Current entity type id.
   */
  private function generateAnonymizationConfigFile(array $entityTypes, string $entityTypeId): void {
    $this->config['database']['anonymization'][$this->table($entityTypes[$entityTypeId]['base table'])] = [
      'where' => [
        'uri' => ['!^private://.*!', 'regex'],
      ],
      'fields' => [
        'filename' => [
          'method' => 'Faker::lexify',
          'args' => ['??????????.txt'],
        ],
        'uri' => [
          'method' => 'concat',
          'args' => [
            'private://',
            '%filename',
          ],
        ],
        'filemime' => 'text/plain',
      ],
    ];
  }

  /**
   * Generates config for User entity table anonymization.
   *
   * @param array $entityTypes
   *   Entity types data.
   * @param string $entityTypeId
   *   Current entity type id.
   */
  private function generateAnonymizationConfigUser(array $entityTypes, string $entityTypeId): void {
    $this->config['database']['anonymization'][$this->table($entityTypes[$entityTypeId]['base table'])] = [
      'where' => [
        'uid' => [0, '>'],
      ],
      'fields' => [
        'name' => [
          'method' => 'concat',
          'args' => [
            ['method' => 'Faker::word'],
            '_',
            '%uid',
          ],
        ],
        'pass' => '-',
        'mail' => [
          'method' => 'concat',
          'args' => [
            '%name',
            '@test.test',
          ],
        ],
        'init' => '%mail',
        'data' => NULL,
      ],
    ];
  }

  /**
   * Returns the percentage of data to dump for entity type.
   *
   * @param string $entityTypeId
   *   Entity type id.
   *
   * @return float
   *   The number from 0 to 1, where 1 means 100%.
   */
  private function getPercentage(string $entityTypeId) {
    return $this->dataPercentage[$entityTypeId] ?? $this->dataPercentage['default'];
  }

  /**
   * Checks if entity type config should be exported.
   *
   * @param string $entityTypeId
   *   Entity type id.
   *
   * @return bool
   *   TRUE - if entity type config should be exported, or FALSE otherwise.
   */
  private function shouldExportEntity(string $entityTypeId): bool {
    return $this->getPercentage($entityTypeId) >= 0;
  }

  /**
   * Returns prefixed table name.
   *
   * @param string $table
   *   Table name without a prefix.
   *
   * @return string
   *   Table name with a prefix.
   */
  private function table(string $table): string {
    return $this->tablePrefix . $table;
  }

}
